// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "Xml.hpp"

using namespace Dawn;

XmlNode::XmlNode() {
  this->child = nullptr;
  this->value.clear();
}


bool_t Xml::isWhitespace(char_t c) {
  return c == ' ' || c == '\r' || c == '\n' || c == '\t';
}

Xml Xml::load(std::string data) {
  size_t j = 0;
  Xml xml;
  Xml::load(&xml, data, &j);
  return xml;
}

void Xml::load(Xml *xml, std::string data, size_t *j) {
  char_t c;
  int32_t level = 0;
  enum XmlParseState doing = XML_PARSE_STATE_DOING_NOTHING;
  enum XmlParseState doingBeforeComment;
  bool_t insideTag = false;
  std::string buffer = "";
  std::string attrKey = "";
  size_t i = *j;
  struct XmlNode childNode;
  
  while(c = data[i++]) {
    if(insideTag) {
      xml->outerXml += c;
    }
    
    switch(doing) {
      case XML_PARSE_STATE_DOING_NOTHING:
        if(c == '>') continue;
        if(c == '<') {
          // Parsing comment?
          if(data[i] == '!' && data[i+1] == '-' && data[i+2] == '-') {
            doingBeforeComment = doing;
            doing = XML_PARSE_STATE_PARSING_COMMENT;
            i += 3;
          } else if(data[i] == '!' && !insideTag) {
            // Likely <!DOCTYPE ...>
            while((c = data[i++]) != '>') {
              // Nothing needs doing here right now, in future may support doctype
            }
            continue;
          } else if(insideTag) {
            if(data[i] == '/') {
              doing = XML_PARSE_STATE_PARSING_CLOSE;
              continue;
            }
          } else {
            doing = XML_PARSE_STATE_PARSING_TAG_NAME;
            level++;
            insideTag = true;
            xml->outerXml += c;
          }
          continue;
        }

        if(insideTag) xml->innerXml += c;
        if(Xml::isWhitespace(c)) continue;// NEEDS TO GO?
        doing = XML_PARSE_STATE_PARSING_VALUE;
        buffer += c;
        break;
      
      case XML_PARSE_STATE_PARSING_TAG_NAME:
        // Just keep reading until we either hit a space (end of the tag name)
        // or a closing tag value, either / or >
        if(Xml::isWhitespace(c) || c == '>' || c == '/') {
          xml->node = buffer;
          buffer = "";
          if(c == '/') {
            level--;
            insideTag = false;
            doing = XML_PARSE_STATE_PARSING_CLOSE;
          } else {
            doing = c == '>' ? XML_PARSE_STATE_PARSING_VALUE : XML_PARSE_STATE_LOOKING_FOR_ATTRIBUTE;
          }
          continue;
        }
        buffer += c;
        break;


      case XML_PARSE_STATE_LOOKING_FOR_ATTRIBUTE:
        // Look until we hit either the end of a tag, or the attribute itself
        if(Xml::isWhitespace(c) || c == '>' || c == '/' || c == '=') {
          if(c == '>' || c == '/') {
            doing = XML_PARSE_STATE_PARSING_VALUE;
            if(c == '/') {
              level--;
              insideTag = false;
              doing = XML_PARSE_STATE_PARSING_CLOSE;
            }
          } else if(c == '=') {
            doing = XML_PARSE_STATE_LOOKING_FOR_ATTRIBUTE_VALUE;
          } else {
            doing = XML_PARSE_STATE_LOOKING_FOR_ATTRIBUTE;
          }

          if(buffer.size() > 0) {
            attrKey = buffer;
            xml->attributes[buffer] = "";
            buffer = "";
          }
          continue;
        }
        buffer += c;
        break;
      
      case XML_PARSE_STATE_LOOKING_FOR_ATTRIBUTE_VALUE:
        // Keep looking until we find a quote mark
        if(Xml::isWhitespace(c)) continue;
        if(c == '>' || c == '/') {
          doing = XML_PARSE_STATE_DOING_NOTHING;
          insideTag = false;
          continue;
        }

        if(c != '"') continue;
        doing = XML_PARSE_STATE_PARSING_ATTRIBUTE_VALUE;
        break;

      case XML_PARSE_STATE_PARSING_ATTRIBUTE_VALUE:
        // Parse the attribute value until we find a quote mark.
        if(c == '\\') {
          c = data[i++];
        } else if(c == '"') {
          doing = XML_PARSE_STATE_LOOKING_FOR_ATTRIBUTE;
          xml->attributes[attrKey] = buffer;
          buffer = "";
          continue;
        }
        buffer += c;
        break;

      case XML_PARSE_STATE_PARSING_VALUE:
        // Keep parsing child until we find a < for an opening/closing tag.
        if(c == '<' && !(data[i] == '<' || data[i-2] == '<')) {
          if(buffer.size() > 0) {
            childNode.nodeType = XML_NODE_TYPE_TEXT;
            childNode.value = buffer;
            xml->childNodes.push_back(childNode);
          }

          // Are we parsing the close tag, or parsing a child?
          if(data[i] == '/') {
            doing = XML_PARSE_STATE_PARSING_CLOSE;
            xml->textContent = buffer;
            buffer.clear();
            continue;
          } else if(data[i] == '!' && data[i+1] == '-' && data[i+2] == '-') {
            doingBeforeComment = doing;
            doing = XML_PARSE_STATE_PARSING_COMMENT;
            i += 3;
            continue;
          }

          // Parsing child
          i -= 1;

          // @deprecated
          auto child = new Xml();
          Xml::load(child, data, &i);
          xml->children.push_back(child);

          childNode = XmlNode();
          childNode.nodeType = XML_NODE_TYPE_ELEMENT;
          childNode.child = child;
          xml->childNodes.push_back(childNode);

          // Remove last char since we kinda already parsed it.
          xml->innerXml += child->outerXml;
          xml->outerXml = xml->outerXml.substr(0, xml->outerXml.size()-1);
          xml->outerXml += child->outerXml;

          buffer.clear();
          continue;
        }

        xml->innerXml += c;

        if(c == '&') {
          // Handle special characters. First read ahead to nearest semicolon OR
          // nearest closing tag.
          std::string sc;
          while(c = data[i++]) {
            xml->innerXml += c;
            if(c == ';') break;
            if(c == '<') assertUnreachable();//Invalid XML
            sc += c;
          }

          if(sc == "lt") {
            buffer += '<';
          } else if(sc == "gt") {
            buffer += '>';
          } else if(sc == "amp") {
            buffer += '&';
          } else if(sc == "apos") {
            buffer += '\'';
          } else if(sc == "quot") {
            buffer += '"';
          } else if(sc == "nbsp") {
            buffer += ' ';
          } else {
            // Try parse as integer
            if(sc.size() > 1 && sc[0] == '#') {
              int code = std::stoi(sc.substr(1));
              buffer += (char)code;
            } else {
              std::cout << "Unknown Special character: " << sc << std::endl;
              assertUnreachable();
            }
          }
        } else {
          buffer += c;
        }

        break;
      
      case XML_PARSE_STATE_PARSING_CLOSE:
        // Just keep parsing until the tag closer finishes.
        if(c != '>') continue;
        doing = XML_PARSE_STATE_DOING_NOTHING;
        
        //TODO: Return index or something?
        *j = i;
        return;

      case XML_PARSE_STATE_PARSING_COMMENT:
        if(c != '-') continue;
        if(data[i] != '-') continue;
        if(data[i+1] != '>') continue;
        i += 2;
        doing = doingBeforeComment;
        break;

      default:
        break;
    }
  }

  *j = i;
}


std::vector<Xml*> Xml::getChildrenOfType(std::string type) {
  std::vector<Xml*> children;
  auto itChildren = this->children.begin();
  while(itChildren != this->children.end()) {
    auto child = *itChildren;
    if(child->node == type) children.push_back(child);
    ++itChildren;
  }
  return children;
}

Xml * Xml::getFirstChildOfType(std::string type) {
  auto itChildren = this->children.begin();
  while(itChildren != this->children.end()) {
    auto child = *itChildren;
    if(child->node == type) return child;
    ++itChildren;
  }
  return nullptr;
}


Xml::~Xml() {
  auto it = this->children.begin();
  while(it != this->children.end()) {
    delete *it;
    ++it;
  }
}